@@ -4,10 +4,9 @@
 # See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
 # SPDX-License-Identifier: MIT
 
-try:
-	from poxy.utils import *
-except:
-	from utils import *
+"""
+Everything relating to the 'project context' object that describes the project for which the documentation is being generated.
+"""
 
 import os
 import copy
@@ -18,7 +17,9 @@ import datetime
 import shutil
 import tempfile
 from schema import Schema, Or, And, Optional
+from .utils import *
 
+__all__ = []
 
 
 #=======================================================================================================================
@@ -67,7 +68,7 @@ def ValueOrArray(typ, name='', length=None):
 # internal helpers
 #=======================================================================================================================
 
-class _Defaults(object):
+class Defaults(object):
 	inline_namespaces = {
 		r'std::(?:literals::)?(?:chrono|complex|string|string_view)_literals'
 	}
@@ -80,6 +81,8 @@ class _Defaults(object):
 		r'__poxy__' :							1,
 		r'__has_include(...)' :					0,
 		r'__has_attribute(...)' :				0,
+		r'__has_builtin(...)' :					0,
+		r'__has_feature(...)' :					0,
 		r'__has_cpp_attribute(...)' :			999999,
 		r'POXY_IMPLEMENTATION_DETAIL(...)' :	r'POXY_IMPLEMENTATION_DETAIL_IMPL',
 		r'POXY_IGNORE(...)' :					r'',
@@ -227,10 +230,10 @@ class _Defaults(object):
 			r'__cplusplus'								: r'202002L',
 			r'__cpp_aggregate_paren_init'				: 201902,
 			r'__cpp_char8_t'							: 201811,
-			r'__cpp_concepts'							: 201907,
+			r'__cpp_concepts'							: 202002,
 			r'__cpp_conditional_explicit'				: 201806,
 			r'__cpp_consteval'							: 201811,
-			r'__cpp_constexpr'							: 201907,
+			r'__cpp_constexpr'							: 202002,
 			r'__cpp_constexpr_dynamic_alloc'			: 201907,
 			r'__cpp_constexpr_in_decltype'				: 201711,
 			r'__cpp_constinit'							: 201907,
@@ -311,11 +314,45 @@ class _Defaults(object):
 			r'__cpp_using_enum'							: 201907,
 		},
 
-		2023 : dict(),
+		2023 : {
+			r'__cplusplus'									: r'202300L',
+			r'__cpp_constexpr'								: 202110,
+			r'__cpp_explicit_this_parameter'				: 202110,
+			r'__cpp_if_consteval'							: 202106,
+			r'__cpp_lib_adaptor_iterator_pair_constructor'	: 202106,
+			r'__cpp_lib_allocate_at_least'					: 202106,
+			r'__cpp_lib_associative_heterogeneous_erasure'	: 202110,
+			r'__cpp_lib_byteswap'							: 202110,
+			r'__cpp_lib_constexpr_typeinfo'					: 202106,
+			r'__cpp_lib_format'								: 202110,
+			r'__cpp_lib_invoke_r'							: 202106,
+			r'__cpp_lib_is_scoped_enum'						: 202011,
+			r'__cpp_lib_monadic_optional'					: 202110,
+			r'__cpp_lib_move_only_function'					: 202110,
+			r'__cpp_lib_optional'							: 202106,
+			r'__cpp_lib_out_ptr'							: 202106,
+			r'__cpp_lib_ranges'								: 202110,
+			r'__cpp_lib_ranges_starts_ends_with'			: 202106,
+			r'__cpp_lib_ranges_zip'							: 202110,
+			r'__cpp_lib_spanstream'							: 202106,
+			r'__cpp_lib_stacktrace'							: 202011,
+			r'__cpp_lib_stdatomic_h'						: 202011,
+			r'__cpp_lib_string_contains'					: 202011,
+			r'__cpp_lib_string_resize_and_overwrite'		: 202110,
+			r'__cpp_lib_to_underlying'						: 202102,
+			r'__cpp_lib_variant'							: 202102,
+			r'__cpp_lib_variant'							: 202106,
+			r'__cpp_multidimensional_subscript'				: 202110,
+			r'__cpp_size_t_suffix'							: 202011,
+		},
 
-		2026 : dict(),
+		2026 : {
+			r'__cplusplus'									: r'202600L',
+		},
 
-		2029 : dict(),
+		2029 : {
+			r'__cplusplus'									: r'202900L',
+		},
 	}
 	autolinks = {
 		# builtins
@@ -478,19 +515,45 @@ class _Defaults(object):
 		r'(?:::)?FVector4s?' : r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FVector4/index.html',
 		r'(?:::)?FVectors?' : r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FVector/index.html',
 		r'(?:::)?TMatrix(?:es|ices)?' : r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/TMatrix/index.html',
+		r'(?:::)?UStaticMesh(?:es)?' : r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/Engine/UStaticMesh/',
+		r'(?:::)?FStaticMeshLODResources' : r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/FStaticMeshLODResources/',
+		r'(?:::)?FRawStaticIndexBuffer(?:s)?' : r'https://docs.unrealengine.com/4.26/en-US/API/Runtime/Engine/FRawStaticIndexBuffer/',
+		r'(?:::)?FStaticMeshVertexBuffers' : r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/FStaticMeshVertexBuffers/',
+		r'(?:::)?FStaticMeshVertexBuffer' : r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/Rendering/FStaticMeshVertexBuffer/',
+		r'(?:::)?FPositionVertexBuffer(?:s)?' : r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/Rendering/FPositionVertexBuffer/',
+		r'(?:::)?TArrayView(?:s)?' : r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Core/Containers/TArrayView/',
+		r'(?:::)?TArray(?:s)?' : r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Core/Containers/TArray/',
 	}
 	navbar = [r'files', r'groups', r'namespaces', r'classes']
 	aliases = {
 		# poxy
 		r'cpp' : r'@code{.cpp}',
-		r'ecpp' : r'@endcode',
-		r'endcpp' : r'@endcode',
+			r'ecpp' : r'@endcode',
+			r'endcpp' : r'@endcode',
 		r'out' : r'@code{.shell-session}',
-		r'eout' : r'@endcode',
-		r'endout' : r'@endcode',
+			r'eout' : r'@endcode',
+			r'endout' : r'@endcode',
+		r'python' : r'@code{.py}',
+			r'epython' : r'@endcode',
+			r'endpython' : r'@endcode',
+		r'meson' : r'@code{.py}',
+			r'emeson' : r'@endcode',
+			r'endmeson' : r'@endcode',
+		r'cmake' : r'@code{.cmake}',
+			r'ecmake' : r'@endcode',
+			r'endcmake' : r'@endcode',
+		r'javascript' : r'@code{.js}',
+			r'ejavascript' : r'@endcode',
+			r'endjavascript' : r'@endcode',
+		r'json' : r'@code{.js}',
+			r'ejson' : r'@endcode',
+			r'endjson' : r'@endcode',
+		r'shell' : r'@code{.shell-session}',
+			r'eshell' : r'@endcode',
+			r'endshell' : r'@endcode',
 		r'bash' : r'@code{.sh}',
-		r'ebash' : r'@endcode',
-		r'endbash' : r'@endcode',
+			r'ebash' : r'@endcode',
+			r'endbash' : r'@endcode',
 		r'detail' : r'@details',
 		r'inline_subheading{1}' : r'[h4]\1[/h4]',
 		r'conditional_return{1}' : r'<strong><em>\1:</em></strong> ',
@@ -588,7 +651,7 @@ class _Defaults(object):
 
 
 
-def _extract_kvps(config, table,
+def extract_kvps(config, table,
 		key_getter=str,
 		value_getter=str,
 		strip_keys=True,
@@ -631,17 +694,17 @@ def _extract_kvps(config, table,
 
 
 
-def _assert_no_unexpected_keys(raw, validated, prefix=''):
+def assert_no_unexpected_keys(raw, validated, prefix=''):
 	for key in raw:
 		if key not in validated:
 			raise Error(rf"Unknown config property '{prefix}{key}'")
 		if isinstance(validated[key], dict):
-			_assert_no_unexpected_keys(raw[key], validated[key], prefix=rf'{prefix}{key}.')
+			assert_no_unexpected_keys(raw[key], validated[key], prefix=rf'{prefix}{key}.')
 	return validated
 
 
 
-class _Warnings(object):
+class Warnings(object):
 	schema = {
 		Optional(r'enabled') 			: bool,
 		Optional(r'treat_as_errors')	: bool,
@@ -669,7 +732,7 @@ class _Warnings(object):
 
 
 
-class _CodeBlocks(object):
+class CodeBlocks(object):
 	schema = {
 		Optional(r'types') 				: ValueOrArray(str, name=r'types'),
 		Optional(r'macros') 			: ValueOrArray(str, name=r'macros'),
@@ -680,12 +743,12 @@ class _CodeBlocks(object):
 	}
 
 	def __init__(self, config, macros):
-		self.types = copy.deepcopy(_Defaults.cb_types)
-		self.macros = copy.deepcopy(_Defaults.cb_macros)
-		self.string_literals = copy.deepcopy(_Defaults.cb_string_literals)
-		self.numeric_literals = copy.deepcopy(_Defaults.cb_numeric_literals)
-		self.enums = copy.deepcopy(_Defaults.cb_enums)
-		self.namespaces = copy.deepcopy(_Defaults.cb_namespaces)
+		self.types = copy.deepcopy(Defaults.cb_types)
+		self.macros = copy.deepcopy(Defaults.cb_macros)
+		self.string_literals = copy.deepcopy(Defaults.cb_string_literals)
+		self.numeric_literals = copy.deepcopy(Defaults.cb_numeric_literals)
+		self.enums = copy.deepcopy(Defaults.cb_enums)
+		self.namespaces = copy.deepcopy(Defaults.cb_namespaces)
 
 		if 'code_blocks' in config:
 			config = config['code_blocks']
@@ -736,7 +799,7 @@ class _CodeBlocks(object):
 
 
 
-class _Inputs(object):
+class Inputs(object):
 	schema = {
 		Optional(r'paths')				: ValueOrArray(str, name=r'paths'),
 		Optional(r'recursive_paths')	: ValueOrArray(str, name=r'recursive_paths'),
@@ -756,11 +819,11 @@ class _Inputs(object):
 			key = r'recursive_paths' if recursive else r'paths'
 			paths = []
 			if not recursive and additional_inputs is not None:
-				paths = paths + [p for p in coerce_collection(additional_inputs)]
+				paths = paths + [p for p in coerce_collection(additional_inputs) if p is not None]
 			if recursive and additional_recursive_inputs is not None:
-				paths = paths + [p for p in coerce_collection(additional_recursive_inputs)]
+				paths = paths + [p for p in coerce_collection(additional_recursive_inputs) if p is not None]
 			if config is not None and key in config:
-				paths = paths + [p for p in coerce_collection(config[key])]
+				paths = paths + [p for p in coerce_collection(config[key]) if p is not None]
 			paths = [p for p in paths if p]
 			paths = [str(p).strip().replace('\\', '/') for p in paths]
 			paths = [Path(p) for p in paths if p]
@@ -773,7 +836,7 @@ class _Inputs(object):
 					raise Error(rf"{key}: '{path}' was not a directory or file")
 				all_paths.add(path)
 				if recursive and path.is_dir():
-					for subdir in enum_subdirs(path, filter=lambda p: not p.name.startswith(r'.'), recursive=True):
+					for subdir in enumerate_directories(path, filter=lambda p: not p.name.startswith(r'.'), recursive=True):
 						all_paths.add(subdir)
 
 		ignores = set()
@@ -789,8 +852,8 @@ class _Inputs(object):
 
 
 
-class _FilteredInputs(_Inputs):
-	schema = combine_dicts(_Inputs.schema, {
+class FilteredInputs(Inputs):
+	schema = combine_dicts(Inputs.schema, {
 		Optional(r'patterns')			: ValueOrArray(str, name=r'patterns')
 	})
 
@@ -817,14 +880,14 @@ class _FilteredInputs(_Inputs):
 
 
 
-class _Sources(_FilteredInputs):
-	schema = combine_dicts(_FilteredInputs.schema, {
+class Sources(FilteredInputs):
+	schema = combine_dicts(FilteredInputs.schema, {
 		Optional(r'strip_paths')		: ValueOrArray(str, name=r'strip_paths'),
 		Optional(r'strip_includes')		: ValueOrArray(str, name=r'strip_includes'),
 		Optional(r'extract_all')		: bool
 	})
 
-	def __init__(self, config, key, input_dir, additional_inputs=None, additional_recursive_inputs=None):
+	def __init__(self, config, key, input_dir, additional_inputs=None, additional_recursive_inputs=None, additional_strip_paths=None):
 		super().__init__(
 			config,
 			key,
@@ -837,15 +900,23 @@ class _Sources(_FilteredInputs):
 		self.strip_includes = []
 		self.extract_all = None
 		if self.patterns is None:
-			self.patterns = copy.deepcopy(_Defaults.source_patterns)
+			self.patterns = copy.deepcopy(Defaults.source_patterns)
 
 		if key not in config:
 			return
 		config = config[key]
 
-		if r'strip_paths' in config:
-			for s in coerce_collection(config[r'strip_paths']):
-				path = s.strip()
+		strip_path_sources = (
+			 coerce_collection(config[r'strip_paths']) if r'strip_paths' in config else None,
+			 coerce_collection(additional_strip_paths) if additional_strip_paths is not None else None
+		)
+		for sps in strip_path_sources:
+			if sps is None:
+				continue
+			for s in sps:
+				if s is None:
+					continue
+				path = str(s).strip()
 				if path:
 					self.strip_paths.append(path)
 
@@ -865,6 +936,7 @@ class _Sources(_FilteredInputs):
 # project context
 #=======================================================================================================================
 
+__all__.append(r'Context')
 class Context(object):
 	"""
 	The context object passed around during one invocation.
@@ -879,16 +951,17 @@ class Context(object):
 			Optional(r'author')					: str,
 			Optional(r'autolinks')				: {str : str},
 			Optional(r'badges')					: {str : FixedArrayOf(str, 2, name=r'badges') },
-			Optional(r'code_blocks')			: _CodeBlocks.schema,
+			Optional(r'changelog')				: Or(str, bool),
+			Optional(r'code_blocks')			: CodeBlocks.schema,
 			Optional(r'cpp')					: Or(str, int, error=r'cpp: expected string or integer'),
 			Optional(r'defines')				: {str : Or(str, int, bool)}, # legacy
 			Optional(r'description')			: str,
-			Optional(r'examples')				: _FilteredInputs.schema,
+			Optional(r'examples')				: FilteredInputs.schema,
 			Optional(r'extra_files')			: ValueOrArray(str, name=r'extra_files'),
 			Optional(r'favicon')				: str,
 			Optional(r'generate_tagfile')		: bool,
 			Optional(r'github')					: str,
-			Optional(r'images')					: _Inputs.schema,
+			Optional(r'images')					: Inputs.schema,
 			Optional(r'implementation_headers') : {str : ValueOrArray(str)},
 			Optional(r'inline_namespaces')		: ValueOrArray(str, name=r'inline_namespaces'),
 			Optional(r'internal_docs')			: bool,
@@ -903,11 +976,11 @@ class Context(object):
 			Optional(r'robots')					: bool,
 			Optional(r'scripts')				: ValueOrArray(str, name=r'scripts'),
 			Optional(r'show_includes')			: bool,
-			Optional(r'sources')				: _Sources.schema,
+			Optional(r'sources')				: Sources.schema,
 			Optional(r'stylesheets')			: ValueOrArray(str, name=r'stylesheets'),
 			Optional(r'tagfiles')				: {str : str},
 			Optional(r'theme')					: Or(r'dark', r'light', r'custom'),
-			Optional(r'warnings')				: _Warnings.schema,
+			Optional(r'warnings')				: Warnings.schema,
 		},
 		ignore_extra_keys=True
 	)
@@ -1030,7 +1103,7 @@ class Context(object):
 		self.verbose_logger = logger if self.__verbose else None
 
 		self.version = lib_version()
-		self.version_string = r'.'.join(self.version)
+		self.version_string = r'.'.join([str(v) for v in lib_version()])
 		if not self.dry_run or self.__verbose:
 			self.info(rf'Poxy v{self.version_string}')
 
@@ -1041,12 +1114,12 @@ class Context(object):
 		threads = int(threads)
 		if threads <= 0:
 			threads = os.cpu_count()
-		self.threads = max(1, min(32, os.cpu_count(), threads))
+		self.threads = max(1, min(os.cpu_count(), threads))
 		self.verbose_value(r'Context.threads', self.threads)
 
 		self.fixers = None
 		self.tagfile_path = None
-		self.warnings = _Warnings(None) # overwritten after reading config; this is for correct 'treat_as_errors' behaviour if we add any pre-config warnings
+		self.warnings = Warnings(None) # overwritten after reading config; this is for correct 'treat_as_errors' behaviour if we add any pre-config warnings
 		if treat_warnings_as_errors:
 			self.warnings.treat_as_errors = True
 
@@ -1116,6 +1189,7 @@ class Context(object):
 			self.input_dir = input_dir
 			self.verbose_value(r'Context.input_dir', self.input_dir)
 			assert_existing_directory(self.input_dir)
+			assert self.input_dir.is_absolute()
 
 			assert self.doxyfile_path is not None
 			self.doxyfile_path = self.doxyfile_path.resolve()
@@ -1125,8 +1199,22 @@ class Context(object):
 				self.config_path = self.config_path.resolve()
 			self.verbose_value(r'Context.config_path', self.config_path)
 			self.verbose_value(r'Context.doxyfile_path', self.doxyfile_path)
+			assert self.config_path.is_absolute()
+			assert self.doxyfile_path.is_absolute()
+
+			# output paths
+			self.xml_dir = Path(self.output_dir, r'xml')
+			self.verbose_value(r'Context.xml_dir', self.xml_dir)
+			self.html_dir = Path(self.output_dir, r'html')
+			self.verbose_value(r'Context.html_dir', self.html_dir)
+			assert self.output_dir.is_absolute()
+			assert self.xml_dir.is_absolute()
+			assert self.html_dir.is_absolute()
+
+			# blog dir
 			self.blog_dir = Path(self.input_dir, r'blog')
 			self.verbose_value(r'Context.blog_dir', self.blog_dir)
+			assert self.blog_dir.is_absolute()
 
 			# temp dirs
 			self.global_temp_dir = Path(tempfile.gettempdir(), r'poxy')
@@ -1141,14 +1229,9 @@ class Context(object):
 			self.verbose_value(r'Context.temp_dir', self.temp_dir)
 			self.temp_pages_dir = Path(self.temp_dir, r'pages')
 			self.verbose_value(r'Context.temp_pages_dir', self.temp_pages_dir)
-
-			# output paths
-			self.xml_dir = Path(self.output_dir, r'xml')
-			self.verbose_value(r'Context.xml_dir', self.xml_dir)
-			self.html_dir = Path(self.output_dir, r'html')
-			self.verbose_value(r'Context.html_dir', self.html_dir)
-			self.mcss_conf_path = Path(self.temp_dir, r'conf.py')
-			self.verbose_value(r'Context.mcss_conf_path', self.mcss_conf_path)
+			assert self.global_temp_dir.is_absolute()
+			assert self.temp_dir.is_absolute()
+			assert self.temp_pages_dir.is_absolute()
 
 			# doxygen
 			if doxygen_path is not None:
@@ -1162,29 +1245,37 @@ class Context(object):
 					if not p.exists() or not p.is_file() or not os.access(str(p), os.X_OK):
 						raise Error(rf'Could not find Doxygen executable in {doxygen_path}')
 					doxygen_path = p
-				assert_existing_file(doxygen_path)
-				self.doxygen_path = doxygen_path
 			else:
-				self.doxygen_path = shutil.which(r'doxygen')
-				if self.doxygen_path is None:
+				doxygen_path = shutil.which(r'doxygen')
+				if doxygen_path is None:
 					raise Error(rf'Could not find Doxygen on system path')
-			if not os.access(str(self.doxygen_path), os.X_OK):
+			assert doxygen_path is not None
+			doxygen_path = coerce_path(doxygen_path).resolve()
+			assert_existing_file(doxygen_path)
+			if not os.access(str(doxygen_path), os.X_OK):
 				raise Error(rf'{doxygen_path} was not an executable file')
+			self.doxygen_path = doxygen_path
 			self.verbose_value(r'Context.doxygen_path', self.doxygen_path)
+			assert self.doxygen_path.is_absolute()
 
 			# m.css
 			if mcss_dir is None:
-				mcss_dir = Path(self.data_dir, r'mcss')
+				mcss_dir = Path(self.data_dir, r'm.css')
 			mcss_dir = coerce_path(mcss_dir).resolve()
 			assert_existing_directory(mcss_dir)
 			assert_existing_file(Path(mcss_dir, r'documentation/doxygen.py'))
 			self.mcss_dir = mcss_dir
 			self.verbose_value(r'Context.mcss_dir', self.mcss_dir)
+			self.mcss_conf_path = Path(self.temp_dir, r'conf.py')
+			self.verbose_value(r'Context.mcss_conf_path', self.mcss_conf_path)
+			assert self.mcss_conf_path.is_absolute()
+			assert self.mcss_dir.is_absolute()
 
 			# misc
-			self.cppref_tagfile = coerce_path(self.data_dir, r'cppreference-doxygen-web.tag.xml')
+			self.cppref_tagfile = coerce_path(self.data_dir, r'cppreference-doxygen-web.tag.xml').resolve()
 			self.verbose_value(r'Context.cppref_tagfile', self.cppref_tagfile)
 			assert_existing_file(self.cppref_tagfile)
+			assert self.cppref_tagfile.is_absolute()
 
 			# initialize temp + output dirs
 			if not self.dry_run:
@@ -1208,9 +1299,9 @@ class Context(object):
 			if self.config_path is not None:
 				assert_existing_file(self.config_path)
 				config = pytomlpp.loads(read_all_text_from_file(self.config_path, logger=self.verbose_logger if dry_run else self.logger))
-			config = _assert_no_unexpected_keys(config, self.__config_schema.validate(config))
+			config = assert_no_unexpected_keys(config, self.__config_schema.validate(config))
 
-			self.warnings = _Warnings(config) # printed in run.py post-doxyfile
+			self.warnings = Warnings(config) # printed in run.py post-doxyfile
 			if treat_warnings_as_errors:
 				self.warnings.treat_as_errors = True
 
@@ -1271,19 +1362,22 @@ class Context(object):
 
 			# project C++ version
 			# defaults to 'current' cpp year version based on (current year - 2)
-			self.cpp = max(int(self.now.year) - 2, 2011)
-			self.cpp = self.cpp - ((self.cpp - 2011) % 3)
+			# 1998, 2003, *range(2011, 2300, 3)
+			default_cpp_year = max(int(self.now.year) - 2, 2011)
+			default_cpp_year = default_cpp_year - ((default_cpp_year - 2011) % 3)
+			self.cpp = default_cpp_year
 			if 'cpp' in config:
 				self.cpp = str(config['cpp']).lstrip('0 \t').rstrip()
 				if not self.cpp:
-					self.cpp = 20
+					self.cpp = default_cpp_year
 				self.cpp = int(self.cpp)
 				if self.cpp in (1998, 98):
 					self.cpp = 1998
 				else:
-					self.cpp = self.cpp % 2000
-					if self.cpp in (3, 11, 14, 17, 20, 23, 26, 29):
-						self.cpp = self.cpp + 2000
+					if self.cpp > 2000:
+						self.cpp -= 2000
+					if self.cpp in [3, *range(11, 300, 3)]:
+						self.cpp += 2000
 					else:
 						raise Error(rf"cpp: '{config['cpp']}' is not a valid cpp standard version")
 			self.verbose_value(r'Context.cpp', self.cpp)
@@ -1371,17 +1465,61 @@ class Context(object):
 						raise Error(rf"failed to parse date from blog post filename '{f.name}': {str(exc)}")
 			self.verbose_value(r'Context.blog_files', self.blog_files)
 
+			# changelog
+			self.changelog = ''
+			if r'changelog' in config:
+				if isinstance(config['changelog'], bool):
+					if config['changelog']:
+						candidate_names = (
+							r'CHANGELOG.md',
+							r'CHANGELOG.txt',
+							r'CHANGELOG',
+							r'HISTORY.md',
+							r'HISTORY.txt',
+							r'HISTORY'
+						)
+						candidate_dir = self.input_dir
+						while True:
+							for name in candidate_names:
+								candidate_file = Path(candidate_dir, name)
+								if candidate_file.exists() and candidate_file.is_file() and candidate_file.stat().st_size <= 1024 * 1024 * 2:
+									self.changelog = candidate_file
+									break
+							if self.changelog or candidate_dir.parent == candidate_dir:
+								break
+							candidate_dir = candidate_dir.parent
+						if not self.changelog and not self.dry_run:
+							self.warning(rf'changelog: Option was set to true but no file with a known changelog file name could be found! Consider using an explicit path.')
+
+				else:
+					self.changelog = coerce_path(config['changelog'])
+					if not self.changelog.is_absolute():
+						self.changelog = Path(self.input_dir, self.changelog)
+					if not self.changelog.exists() or not self.changelog.is_file():
+						raise Error(rf'changelog: {config["changelog"]} did not exist or was not a file')
+			self.verbose_value(r'Context.changelog', self.changelog)
+			if self.changelog and not self.dry_run:
+				copy_file(self.changelog, Path(self.temp_pages_dir, r'poxy_changelog.md'), logger=self.verbose_logger)
+				self.changelog = Path(self.temp_pages_dir, r'poxy_changelog.md')
+
 			# sources (INPUT, FILE_PATTERNS, STRIP_FROM_PATH, STRIP_FROM_INC_PATH, EXTRACT_ALL)
-			self.sources = _Sources(
+			self.sources = Sources(
 				config,
 				r'sources',
 				self.input_dir,
-				additional_inputs=[ self.temp_pages_dir, *[f for f,d in self.blog_files] ]
+				additional_inputs=(
+					self.temp_pages_dir if not self.dry_run else None,
+					self.changelog if self.changelog and not self.dry_run else None,
+					*[f for f,d in self.blog_files]
+				),
+				additional_strip_paths=(
+					self.temp_pages_dir if not self.dry_run else None,
+				)
 			)
 			self.verbose_object(r'Context.sources', self.sources)
 
 			# images (IMAGE_PATH)
-			self.images = _Inputs(
+			self.images = Inputs(
 				config,
 				r'images',
 				self.input_dir,
@@ -1390,7 +1528,7 @@ class Context(object):
 			self.verbose_object(r'Context.images', self.images)
 
 			# examples (EXAMPLES_PATH, EXAMPLE_PATTERNS)
-			self.examples = _FilteredInputs(
+			self.examples = FilteredInputs(
 				config,
 				r'examples',
 				self.input_dir,
@@ -1403,7 +1541,7 @@ class Context(object):
 				self.cppref_tagfile : (self.cppref_tagfile, r'http://en.cppreference.com/w/')
 			}
 			self.unresolved_tagfiles = False
-			for k,v in _extract_kvps(config, 'tagfiles').items():
+			for k,v in extract_kvps(config, 'tagfiles').items():
 				source = str(k)
 				dest = str(v)
 				if source and dest:
@@ -1430,7 +1568,7 @@ class Context(object):
 					if val:
 						self.navbar.append(val)
 			else:
-				self.navbar = copy.deepcopy(_Defaults.navbar)
+				self.navbar = copy.deepcopy(Defaults.navbar)
 			for i in range(len(self.navbar)):
 				if self.navbar[i] == 'classes':
 					self.navbar[i] = 'annotated'
@@ -1443,7 +1581,7 @@ class Context(object):
 
 			# <meta> tags
 			self.meta_tags = {}
-			for k, v in _extract_kvps(config, 'meta_tags', allow_blank_values=True).items():
+			for k, v in extract_kvps(config, 'meta_tags', allow_blank_values=True).items():
 				self.meta_tags[k] = v
 			self.verbose_value(r'Context.meta_tags', self.meta_tags)
 
@@ -1454,7 +1592,7 @@ class Context(object):
 			self.verbose_value(r'Context.robots', self.robots)
 
 			# inline namespaces for old versions of doxygen
-			self.inline_namespaces = copy.deepcopy(_Defaults.inline_namespaces)
+			self.inline_namespaces = copy.deepcopy(Defaults.inline_namespaces)
 			if 'inline_namespaces' in config:
 				for ns in coerce_collection(config['inline_namespaces']):
 					namespace = ns.strip()
@@ -1510,26 +1648,30 @@ class Context(object):
 			self.verbose_value(r'Context.favicon', self.favicon)
 
 			# macros (PREDEFINED)
-			self.macros = copy.deepcopy(_Defaults.macros)
+			self.macros = copy.deepcopy(Defaults.macros)
 			for s in (r'defines', r'macros'):
-				for k, v in _extract_kvps(config, s,
+				for k, v in extract_kvps(config, s,
 						value_getter=lambda v: (r'true' if v else r'false') if isinstance(v, bool) else str(v),
 						allow_blank_values=True,
 					).items():
 					self.macros[k] = v
 			non_cpp_def_macros = copy.deepcopy(self.macros)
 			cpp_defs = dict()
-			for ver in (1998, 2003, 2011, 2014, 2017, 2020, 2023, 2026, 2029):
+			for ver in [1998, 2003, *range(2011, 2300, 3)]:
 				if ver > self.cpp:
 					break
-				for k, v in _Defaults.cpp_builtin_macros[ver].items():
+				if ver not in Defaults.cpp_builtin_macros:
+					continue
+				for k, v in Defaults.cpp_builtin_macros[ver].items():
 					cpp_defs[k] = v
-			for k, v in cpp_defs.items():
+			cpp_defs = [(k, v) for k, v in cpp_defs.items()]
+			cpp_defs.sort(key=lambda kvp: k[0])
+			for k, v in cpp_defs:
 				self.macros[k] = v
 			self.verbose_value(r'Context.macros', self.macros)
 
 			# autolinks
-			self.autolinks = [(k, v) for k, v in _Defaults.autolinks.items()]
+			self.autolinks = [(k, v) for k, v in Defaults.autolinks.items()]
 			if 'autolinks' in config:
 				for pattern, u in config['autolinks'].items():
 					uri = u.strip()
@@ -1545,7 +1687,7 @@ class Context(object):
 			self.verbose_value(r'Context.autolinks', self.autolinks)
 
 			# aliases (ALIASES)
-			self.aliases = copy.deepcopy(_Defaults.aliases)
+			self.aliases = copy.deepcopy(Defaults.aliases)
 			if 'aliases' in config:
 				for k, v in config['aliases'].items():
 					alias = k.strip()
@@ -1594,7 +1736,7 @@ class Context(object):
 			self.verbose_value(r'Context.extra_files', self.extra_files)
 
 			# code_blocks
-			self.code_blocks = _CodeBlocks(config, non_cpp_def_macros) # printed in run.py post-xml
+			self.code_blocks = CodeBlocks(config, non_cpp_def_macros) # printed in run.py post-xml
 
 		# initialize other data from files on disk
 		self.__init_data_files(self)
