@@ -4,27 +4,24 @@
 # See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
 # SPDX-License-Identifier: MIT
 
-try:
-	from poxy.utils import *
-	import poxy.project as project
-	import poxy.doxygen as doxygen
-	import poxy.soup as soup
-	import poxy.fixers as fixers
-except:
-	from utils import *
-	import project
-	import doxygen
-	import soup
-	import fixers
+"""
+The 'actually do the thing' module.
+"""
 
 import os
 import subprocess
 import concurrent.futures as futures
-
 import tempfile
 import requests
 from lxml import etree
 from io import BytesIO, StringIO
+from .utils import *
+from . import project
+from . import doxyfile
+from . import soup
+from . import fixers
+
+__all__ = []
 
 
 #=======================================================================================================================
@@ -131,6 +128,7 @@ _doxygen_overrides  = (
 		(r'STRICT_PROTO_MATCHING',	False),
 		(r'SUBGROUPING', 			True),
 		(r'TAB_SIZE',				4),
+		(r'TOC_INCLUDE_HEADINGS',	3),
 		(r'TYPEDEF_HIDES_STRUCT',	False),
 		(r'UML_LOOK',				False),
 		(r'USE_HTAGS',				False),
@@ -145,11 +143,11 @@ _doxygen_overrides  = (
 
 
 
-def _preprocess_doxyfile(context):
+def preprocess_doxyfile(context):
 	assert context is not None
 	assert isinstance(context, project.Context)
 
-	with doxygen.Doxyfile(
+	with doxyfile.Doxyfile(
 			doxyfile_path = context.doxyfile_path,
 			cwd = context.input_dir,
 			logger = context.verbose_logger,
@@ -219,7 +217,7 @@ def _preprocess_doxyfile(context):
 			else:
 				df.set_value(r'GENERATE_TAGFILE', None)
 
-			df.set_value(r'NUM_PROC_THREADS', context.threads)
+			df.set_value(r'NUM_PROC_THREADS', min(context.threads, 32))
 			df.add_value(r'CLANG_OPTIONS', rf'-std=c++{context.cpp%100}')
 			df.add_value(r'CLANG_OPTIONS', r'-Wno-everything')
 
@@ -353,6 +351,8 @@ def _preprocess_doxyfile(context):
 				if context.github:
 					footer.append(rf'<a href="https://github.com/{context.github}/">Github</a>')
 					footer.append(rf'<a href="https://github.com/{context.github}/issues">Report an issue</a>')
+				if context.changelog:
+					footer.append(rf'<a href="md_poxy_changelog.html">Changelog</a>')
 				if context.license and context.license[r'uri']:
 					footer.append(rf'<a href="{context.license["uri"]}" target="_blank">License</a>')
 				if context.generate_tagfile:
@@ -378,29 +378,22 @@ def _preprocess_doxyfile(context):
 			context.info(rf'# generated by Poxy v{context.version_string}')
 			context.info(r'#====================================================================================')
 			context.info(df.get_text())
-			context.info(r'##! ---------------------------------------------------------------------------------')
-			context.info(r'##! m.css conf.py:')
-			context.info(r'##! ---------------------------------------------------------------------------------')
-			context.info(conf_py.getvalue(), indent=r'##! ')
+			context.info(r'## ---------------------------------------------------------------------------------')
+			context.info(r'## m.css conf.py:')
+			context.info(r'## ---------------------------------------------------------------------------------')
+			context.info(conf_py.getvalue(), indent=r'## ')
 			context.info(r'#====================================================================================')
 		else:
 			context.verbose(r'Effective Doxyfile:')
 			context.verbose(df.get_text(), indent=r'    ')
-			context.verbose(r'    ##! --------------------------------------------------------------------------')
-			context.verbose(r'    ##! m.css conf.py:')
-			context.verbose(r'    ##! --------------------------------------------------------------------------')
-			context.verbose(conf_py.getvalue(), indent=r'    ##! ')
+			context.verbose(r'    ## --------------------------------------------------------------------------')
+			context.verbose(r'    ## m.css conf.py:')
+			context.verbose(r'    ## --------------------------------------------------------------------------')
+			context.verbose(conf_py.getvalue(), indent=r'    ## ')
 
 
-def _filter_filenames(files, include, exclude):
-	if include is not None:
-		files = [f for f in files if include.search(f.name) is not None]
-	if exclude is not None:
-		files = [f for f in files if exclude.search(f.name) is None]
-	return files
 
-
-def _postprocess_xml(context):
+def postprocess_xml(context):
 	assert context is not None
 	assert isinstance(context, project.Context)
 
@@ -422,7 +415,7 @@ def _postprocess_xml(context):
 
 		inline_namespace_ids = None
 		if context.inline_namespaces:
-			inline_namespace_ids = [f'namespace{doxygen.mangle_name(ns)}' for ns in context.inline_namespaces]
+			inline_namespace_ids = [f'namespace{doxyfile.mangle_name(ns)}' for ns in context.inline_namespaces]
 
 		implementation_header_data = None
 		implementation_header_mappings = None
@@ -435,8 +428,8 @@ def _postprocess_xml(context):
 				(
 					hp,
 					os.path.basename(hp),
-					doxygen.mangle_name(os.path.basename(hp)),
-					[(i, os.path.basename(i), doxygen.mangle_name(os.path.basename(i))) for i in impl]
+					doxyfile.mangle_name(os.path.basename(hp)),
+					[(i, os.path.basename(i), doxyfile.mangle_name(os.path.basename(i))) for i in impl]
 				)
 				for hp, impl in context.implementation_headers
 			]
@@ -462,7 +455,7 @@ def _postprocess_xml(context):
 			# pre-pass to delete junk files
 			if 1:
 				# 'file' entries for markdown and dox files
-				dox_files = [rf'*{doxygen.mangle_name(ext)}.xml' for ext in (r'.dox', r'.md')]
+				dox_files = [rf'*{doxyfile.mangle_name(ext)}.xml' for ext in (r'.dox', r'.md')]
 				dox_files.append(r'md_home.xml')
 				for xml_file in get_all_files(context.xml_dir, any=dox_files):
 					delete_file(xml_file, logger=context.verbose_logger)
@@ -471,7 +464,7 @@ def _postprocess_xml(context):
 				deleted = True
 				while deleted:
 					deleted = False
-					for xml_file in get_all_files(context.xml_dir, all=(r'dir_*.xml')):
+					for xml_file in get_all_files(context.xml_dir, all=(r'dir*.xml')):
 						xml = etree.parse(str(xml_file), parser=xml_parser)
 						compounddef = xml.getroot().find(r'compounddef')
 						if compounddef is None or compounddef.get(r'kind') != r'dir':
@@ -486,6 +479,22 @@ def _postprocess_xml(context):
 							delete_file(xml_file, logger=context.verbose_logger)
 							deleted = True
 
+				# concepts - not currently supported by m.css
+				if not context.xml_only:
+					for xml_file in get_all_files(context.xml_dir, all=(r'concept*.xml')):
+						xml = etree.parse(str(xml_file), parser=xml_parser)
+						compounddef = xml.getroot().find(r'compounddef')
+						if compounddef is None or compounddef.get(r'kind') != r'concept':
+							continue
+						compoundname = compounddef.find(r'compoundname')
+						assert compoundname is not None
+						assert compoundname.text
+						context.warning(
+							rf"C++20 concepts are not currently supported! No documentation will be generated for '{compoundname.text}'."
+							+ r" Surround your concepts in a '@cond poxy_supports_concepts' block to suppress this warning until"
+							+ r" poxy is updated to support them.")
+						delete_file(xml_file, logger=context.verbose_logger)
+
 			extracted_implementation = False
 			tentative_macros = regex_or(context.code_blocks.macros)
 			macros = set()
@@ -507,7 +516,7 @@ def _postprocess_xml(context):
 				if root.tag == r'doxygenindex':
 
 					# remove entries for files we might have explicitly deleted above
-					for compound in [tag for tag in root.findall(r'compound') if tag.get(r'kind') in (r'file', r'dir')]:
+					for compound in [tag for tag in root.findall(r'compound') if tag.get(r'kind') in (r'file', r'dir', r'concept')]:
 						ref_file = Path(context.xml_dir, rf'{compound.get(r"refid")}.xml')
 						if not ref_file.exists():
 							root.remove(compound)
@@ -638,36 +647,37 @@ def _postprocess_xml(context):
 							# fix functions where keywords like 'friend' have been erroneously included in the return type
 							if 1:
 								members = [m for m in section.findall(r'memberdef') if m.get(r'kind') in (r'friend', r'function')]
-								keywords_behaving_badly = (
-									(r'constexpr ', r'constexpr', r'yes'),
-									(r'consteval ', r'constexpr', r'yes'),
-									(r'explicit ', r'explicit', r'yes'),
-									(r'static ', r'static', r'yes'),
-									(r'friend ', None, None),
-									(r'inline ', r'inline', r'yes'),
+								attribute_keywords = (
+									(r'constexpr', r'constexpr', r'yes'),
+									(r'consteval', r'consteval', r'yes'),
+									(r'explicit', r'explicit', r'yes'),
+									(r'static', r'static', r'yes'),
+									(r'friend', None, None),
+									(r'inline', r'inline', r'yes'),
+									(r'virtual', r'virt', r'virtual')
 								)
 								for member in members:
 									type = member.find(r'type')
 									if type is None or type.text is None:
 										continue
 									matched_bad_keyword = True
-									removed_constexpr = False
 									while matched_bad_keyword:
 										matched_bad_keyword = False
-										for kw, attr, attr_value in keywords_behaving_badly:
-											if type.text.startswith(kw):
-												matched_bad_keyword = True
-												changed = True
-												type.text = type.text[len(kw):]
-												if attr is not None:
-													member.set(attr, attr_value)
-												elif kw == r'friend ':
-													member.set(r'kind', r'friend')
-												if kw == r'constexpr ':
-													removed_constexpr = True
-									# hack: m.css seems to need 'constexpr' to be a part of the type and ignores the xml attribute O_o
-									if removed_constexpr and not context.xml_only:
-										type.text = rf'constexpr {type.text}'
+										for kw, attr, attr_value in attribute_keywords:
+											if type.text == kw: # constructors
+												type.text = ''
+											elif type.text.startswith(kw + ' '):
+												type.text = type.text[len(kw):].strip()
+											elif type.text.endswith(' ' + kw):
+												type.text = type.text[:len(kw)].strip()
+											else:
+												continue
+											matched_bad_keyword = True
+											changed = True
+											if attr is not None:
+												member.set(attr, attr_value)
+											elif kw == r'friend':
+												member.set(r'kind', r'friend')
 
 							# re-sort members to override Doxygen's weird and stupid sorting 'rules'
 							if 1:
@@ -879,7 +889,7 @@ def _initialize_worker(context):
 
 
 
-def _postprocess_html_file(path, context=None):
+def postprocess_html_file(path, context=None):
 	assert path is not None
 	assert isinstance(path, Path)
 	assert path.is_absolute()
@@ -891,32 +901,24 @@ def _postprocess_html_file(path, context=None):
 	assert context is not None
 	assert isinstance(context, project.Context)
 
-	html_fixers = [f for f in context.fixers if isinstance(f, fixers.HTMLFixer)]
-	plain_text_fixers = [f for f in context.fixers if isinstance(f, fixers.PlainTextFixer)]
-
-	context.verbose(rf'Post-processing {path}')
-	html_changed = False
-	plain_text_changed = False
-
+	context.info(rf'Post-processing {path}')
+	changed = False
 	try:
-		if html_fixers:
-			doc = soup.HTMLDocument(path, logger=context.verbose_logger)
-			for fix in html_fixers:
+		for fix in context.fixers:
+			if isinstance(fix, fixers.HTMLFixer):
+				doc = soup.HTMLDocument(path, logger=context.verbose_logger)
 				if fix(doc, context):
 					doc.smooth()
-					html_changed = True
-			if html_changed:
-				doc.flush()
-
-		if plain_text_fixers:
-			doc = [ read_all_text_from_file(path, logger=context.verbose_logger) ]
-			for fix in plain_text_fixers:
+					doc.flush()
+					changed = True
+			elif isinstance(fix, fixers.PlainTextFixer):
+				doc = [ read_all_text_from_file(path, logger=context.verbose_logger), path ]
 				if fix(doc, context):
-					plain_text_changed = True
-			if plain_text_changed:
-				context.verbose(rf'Writing {path}')
-				with open(path, 'w', encoding='utf-8', newline='\n') as f:
-					f.write(doc[0])
+					context.verbose(rf'Writing {path}')
+					with open(path, 'w', encoding='utf-8', newline='\n') as f:
+						f.write(doc[0])
+					changed = True
+
 	except Exception as e:
 		context.info(rf'{type(e).__name__} raised while post-processing {path}')
 		raise
@@ -924,15 +926,15 @@ def _postprocess_html_file(path, context=None):
 		context.info(rf'Error occurred while post-processing {path}')
 		raise
 
-	return html_changed or plain_text_changed
+	return changed
 
 
 
-def _postprocess_html(context):
+def postprocess_html(context):
 	assert context is not None
 	assert isinstance(context, project.Context)
 
-	files = _filter_filenames(
+	files = filter_filenames(
 		get_all_files(context.html_dir, any=('*.html', '*.htm')),
 		context.html_include,
 		context.html_exclude
@@ -940,7 +942,7 @@ def _postprocess_html(context):
 	if not files:
 		return
 
-	threads = min(len(files), context.threads, 4)
+	threads = min(len(files), context.threads, 8) # diminishing returns after 8
 
 	with ScopeTimer(rf'Post-processing {len(files)} HTML files', print_start=True, print_end=context.verbose_logger):
 		context.fixers = (
@@ -957,11 +959,12 @@ def _postprocess_html(context):
 			fixers.EmptyTags(),
 			fixers.HeadTags(),
 			fixers.ImplementationDetails(),
+			fixers.MarkdownPages(),
 		)
 		context.verbose(rf'Post-processing {len(files)} HTML files...')
 		if threads > 1:
 			with futures.ProcessPoolExecutor(max_workers=threads, initializer=_initialize_worker, initargs=(context,)) as executor:
-				jobs = [ executor.submit(_postprocess_html_file, file) for file in files ]
+				jobs = [ executor.submit(postprocess_html_file, file) for file in files ]
 				for future in futures.as_completed(jobs):
 					try:
 						future.result()
@@ -974,7 +977,7 @@ def _postprocess_html(context):
 
 		else:
 			for file in files:
-				_postprocess_html_file(file, context)
+				postprocess_html_file(file, context)
 
 
 
@@ -982,7 +985,7 @@ def _postprocess_html(context):
 # RUN
 #=======================================================================================================================
 
-def _read_output_streams(stdout, stderr):
+def read_output_streams(stdout, stderr):
 	stdout.seek(0)
 	stderr.seek(0)
 	return {
@@ -992,7 +995,7 @@ def _read_output_streams(stdout, stderr):
 
 
 
-def _dump_output_streams(context, outputs, source=''):
+def dump_output_streams(context, outputs, source=''):
 	if source:
 		source = rf'{source} '
 	if outputs[r'stdout']:
@@ -1023,7 +1026,7 @@ _warnings_ignored = (
 	r'inline code has multiple lines, fallback to a code block',
 	r'libgs not found'
 )
-def _extract_warnings(outputs):
+def extract_warnings(outputs):
 	if not outputs:
 		return []
 
@@ -1063,6 +1066,7 @@ def _extract_warnings(outputs):
 
 
 
+__all__.append(r'run')
 def run(config_path='.',
 		output_dir='.',
 		threads=-1,
@@ -1095,7 +1099,7 @@ def run(config_path='.',
 	) as context:
 
 		# preprocess the doxyfile
-		_preprocess_doxyfile(context)
+		preprocess_doxyfile(context)
 		context.verbose_object(r'Context.warnings', context.warnings)
 
 		if context.dry_run:
@@ -1120,6 +1124,23 @@ def run(config_path='.',
 
 		make_temp_file = lambda: tempfile.SpooledTemporaryFile(mode='w+', newline='\n', encoding='utf-8')
 
+		# precondition the change log page (at this point it is already a temp copy)
+		if context.changelog:
+			text = read_all_text_from_file(context.changelog, logger=context.verbose_logger).strip()
+			text = text.replace('\r\n', 		'\n')
+			text = text.replace(r'&amp;',		r'__poxy_thiswasan_amp')
+			text = text.replace(r'@',			r'__poxy_thiswasan_at')
+			text = text.replace(r'&#xFE0F;', 	r'__poxy_thiswasan_fe0f')
+			if text.find(r'@tableofcontents') == -1 and text.find('\\tableofcontents') == -1 and text.find('[TOC]') == -1:
+				#text = f'[TOC]\n\n{text}'
+				nlnl = text.find('\n\n')
+				if nlnl != -1:
+					text = f'{text[:nlnl]}\n\n\\tableofcontents\n\n{text[nlnl:]}'
+				pass
+			text += '\n\n'
+			with open(context.changelog, r'w', encoding=r'utf-8', newline='\n') as f:
+				f.write(text)
+
 		# run doxygen to generate the xml
 		if 1:
 			with ScopeTimer(r'Generating XML files with Doxygen', print_start=True, print_end=context.verbose_logger) as t:
@@ -1134,14 +1155,14 @@ def run(config_path='.',
 						)
 					except:
 						context.info(r'Doxygen failed!')
-						_dump_output_streams(context, _read_output_streams(stdout, stderr), source=r'Doxygen')
+						dump_output_streams(context, read_output_streams(stdout, stderr), source=r'Doxygen')
 						raise
 					if context.is_verbose() or context.warnings.enabled:
-						outputs = _read_output_streams(stdout, stderr)
+						outputs = read_output_streams(stdout, stderr)
 						if context.is_verbose():
-							_dump_output_streams(context, outputs, source=r'Doxygen')
+							dump_output_streams(context, outputs, source=r'Doxygen')
 						if context.warnings.enabled:
-							warnings = _extract_warnings(outputs)
+							warnings = extract_warnings(outputs)
 							for w in warnings:
 								context.warning(w)
 
@@ -1155,7 +1176,7 @@ def run(config_path='.',
 
 		# post-process xml files
 		if 1:
-			_postprocess_xml(context)
+			postprocess_xml(context)
 
 		if context.xml_only:
 			return
@@ -1189,14 +1210,14 @@ def run(config_path='.',
 						)
 					except:
 						context.info(r'm.css failed!')
-						_dump_output_streams(context, _read_output_streams(stdout, stderr), source=r'm.css')
+						dump_output_streams(context, read_output_streams(stdout, stderr), source=r'm.css')
 						raise
 					if context.is_verbose() or context.warnings.enabled:
-						outputs = _read_output_streams(stdout, stderr)
+						outputs = read_output_streams(stdout, stderr)
 						if context.is_verbose():
-							_dump_output_streams(context, outputs, source=r'm.css')
+							dump_output_streams(context, outputs, source=r'm.css')
 						if context.warnings.enabled:
-							warnings = _extract_warnings(outputs)
+							warnings = extract_warnings(outputs)
 							for w in warnings:
 								context.warning(w)
 
@@ -1214,4 +1235,4 @@ def run(config_path='.',
 
 		# post-process html files
 		if 1:
-			_postprocess_html(context)
+			postprocess_html(context)
